/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.httpd;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

/**
 * Represents a connection which reads a request and writes a
 * response.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see ConnectionSelector
 * @see Acceptor
 */
class Connection {

   private SocketChannel socketChannel;
   private ByteBuffer requestLineBuffer;
   private ByteBuffer responseLineBuffer;
   private int endOfLineIndex;
   private SelectionKey key;
   private FileChannel fileChannel;
   private long filePos;
   private long fileLength;
   private int httpStatus;
   private String httpMessage;
   private String uri;
   private String protocol;


   /**
    * Initializes a connection with a socketchannel.
    */
   public Connection(SocketChannel socketChannel) throws IOException {
      // turn Nagle's algorithm off
      socketChannel.socket().setTcpNoDelay(true);
      // the channel shall NOT block
      socketChannel.configureBlocking(false);
      requestLineBuffer = ByteBuffer.allocate(512);
      // default http status code: OK
      httpStatus = 200;
      // default http status message
      httpMessage = "OK";
      // default protocol version
      protocol = "HTTP/0.9";
   }

   /**
    * Registers this connection with the given selector.
    */
   public void register(Selector selector) throws IOException {
      key = socketChannel.register(selector, SelectionKey.OP_READ);
      key.attach(this);
   }

   /**
    * Reads a request. If something goes wrong, an errorcode will be set.
    * If the request is read completely, {@link #prepareForResponse()} is
    * called.
    */
   public void readRequest() throws IOException {
      try {
         if (!requestLineBuffer.hasRemaining()) {
            setError(414, "Request URI too long.");
            prepareForResponse();
            return;
         }
         socketChannel.read(requestLineBuffer);
         if (!isRequestLineRead()) {
            return;
         }
         requestLineBuffer.flip();
         byte[] b = new byte[endOfLineIndex];
         requestLineBuffer.get(b);
         String requestline = new String(b, 0);
         StringTokenizer st = new StringTokenizer(requestline, " \r\n");
         String method = st.nextToken();
         uri = st.nextToken();
         File file = new File(uri.substring(1));
         if (st.hasMoreTokens()) {
            protocol = st.nextToken();
         }
         if (!method.equals("GET")) {
            setError(405, "Method " + method + " is not supported.");
         } else if (!file.exists() || file.isDirectory()) {
            setError(404, "Resource " + uri + " was not found.");
         } else if (!file.canRead()) {
            setError(403, "Forbidden: " + uri);
         } else {
            fileLength = file.length();
            fileChannel = new FileInputStream(file).getChannel();
         }
         prepareForResponse();
      } catch (NoSuchElementException nsee) {
         // we couldn't read enough tokens
         setError(400, "Bad request.");
      } catch (Exception e) {
         // something went really wrong
         setError(500, "Internal Server Error.");
         prepareForResponse();
         e.printStackTrace();
      }
   }

   /**
    * Creates a buffer which holds a response line.
    */
   private void prepareForResponse() throws IOException {
      StringBuffer responseLine = new StringBuffer(128);
      // write response line only if Http >= 1.0
      if (!protocol.equals("HTTP/0.9")) {
         responseLine.append("HTTP/1.0 " + httpStatus + " " + httpMessage + "\r\n");
         // in case of an error we don't need headers
         if (httpStatus != 200) {
            responseLine.append("\r\n");
         } else {
            // header for the file
            responseLine.append("Content-Type: " + NIOHttpd.guessContentType(uri) + "\r\n\r\n");
         }
      }
      if (httpStatus != 200) {
         // error message
         responseLine.append("<HTML><BODY><H1>" + httpMessage + "</H1></BODY></HTML>");
      }
      responseLineBuffer = ByteBuffer.wrap(responseLine.toString().getBytes("ASCII"));
      key.interestOps(SelectionKey.OP_WRITE);
      key.selector().wakeup();
   }

   /**
    * Indicates whether the reqestline has been read.
    */
   private boolean isRequestLineRead() {
      for (; endOfLineIndex < requestLineBuffer.limit(); endOfLineIndex++) {
         if (requestLineBuffer.get(endOfLineIndex) == '\r') return true;
      }
      return false;
   }

   /**
    * Writes the responseLineBuffer and (if necessary) the requested file
    * to the client.
    */
   public void writeResponse() throws IOException {
      // zunaechst mal den response buffer schreiben
      if (responseLineBuffer.hasRemaining()) {
         socketChannel.write(responseLineBuffer);
      }
      if (!responseLineBuffer.hasRemaining()) {
         if (httpStatus != 200) {
            close();
         } else {
            filePos += fileChannel.transferTo(filePos, (int) Math.min(64 * 1024, fileLength - filePos), socketChannel);
            if (filePos == fileLength) {
               close();
            }
         }
      }
   }

   /**
    * Sets an error state.
    */
   private void setError(int httpStatus, String httpMessage) {
      this.httpStatus = httpStatus;
      this.httpMessage = httpMessage;
   }

   /**
    * Cancels the selection key and closes all open channels.
    */
   public void close() {
      try {
         if (key != null) key.cancel();
      } catch (Exception e) {
         /* ignore */
      }
      try {
         if (socketChannel != null) socketChannel.close();
      } catch (Exception e) {
         /* ignore */
      }
      try {
         if (fileChannel != null) fileChannel.close();
      } catch (Exception e) {
         /* ignore */
      }
   }
}
