/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.strings;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.text.Collator;
import java.util.*;

/**
 * Compares performance of <code>new TreeMap(String.CASE_INSENSITIVE_ORDER)</code>
 * and <code>CaseInsensitiveCollections.caseInsensitiveMap(new HashMap())</code>.
 * Loads words from a file, add their uppercase version to the maps and
 * then calls get() for the non-uppercased version of the word.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class CaseInsensitiveMapDemo {

    private static Collator collator;

    static {
        collator = Collator.getInstance(Locale.GERMANY);
        collator.setStrength(Collator.SECONDARY); // only capitalisation is ignored
    }

    private static List words = new ArrayList();
    private static Map caseInsensitiveTreeMap = new TreeMap(String.CASE_INSENSITIVE_ORDER);
    private static Map caseInsensitiveHashMap = new CaseInsensitiveMap();
    private static Map caseInsensitiveCollationsKeyHashMap = new CollationKeyMap(Locale.GERMANY, Collator.SECONDARY);
    private static Map caseInsensitiveCollatorTreeMap = new TreeMap(collator);

    private static Map caseSensitiveHashMap = new HashMap();

    public static void main(String[] args) throws IOException {
        if (args.length != 1) usage();
        fillWordList(args[0]);
        fillMapLowerCase(caseInsensitiveHashMap);
        fillMapLowerCase(caseInsensitiveTreeMap);
        fillMapLowerCase(caseInsensitiveCollationsKeyHashMap);
        fillMapLowerCase(caseInsensitiveCollatorTreeMap);
        fillMap(caseSensitiveHashMap);
        int iterations = 500;

        System.out.println("Words: " + words.size());
        System.out.println("CaseInsensitive different Words: " + caseInsensitiveHashMap.size());

        long start;
        long time;

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveHashMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveHashMap) first run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveHashMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveHashMap) second run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveTreeMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveTreeMap) first run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveTreeMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveTreeMap) second run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveCollationsKeyHashMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveCollationsKeyHashMap) first run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveCollationsKeyHashMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveCollationsKeyHashMap) second run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveCollatorTreeMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveCollatorTreeMap) first run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseInsensitiveCollatorTreeMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseInsensitiveCollatorTreeMap) second run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseSensitiveHashMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseSensitiveHashMap) first run: " + time);

        start = System.currentTimeMillis();
        for (int i = 0; i < iterations; i++) getAllWords(caseSensitiveHashMap);
        time = System.currentTimeMillis() - start;
        System.out.println("getAllWords(caseSensitiveHashMap) second run: " + time);
    }

    private static void fillMapLowerCase(Map m) {
        for (Iterator i = words.iterator(); i.hasNext();) {
            m.put(new String((String) i.next()).toUpperCase(), "blub");
        }
    }

    private static void fillMap(Map m) {
        for (Iterator i = words.iterator(); i.hasNext();) {
            m.put(new String((String) i.next()), "blub");
        }
    }

    private static void fillWordList(String fileName) throws IOException {
        BufferedReader in = null;
        try {
            in = new BufferedReader(new FileReader(fileName));
            String word;
            while ((word = in.readLine()) != null) {
                //word = word.toUpperCase();
                //word = word.toUpperCase().toLowerCase();
                words.add(word);
            }
        } finally {
            try {
                in.close();
            } catch (IOException ioe) {
            }
        }
    }

    private static void getAllWords(Map map) {
        for (Iterator i = words.iterator(); i.hasNext();) {
            String s = (String) i.next();
            if (map.get(s) == null) throw new NullPointerException("Didn't find element: " + s);
        }
    }

    public static void usage() {
        System.out.println("java " + CaseInsensitiveMapDemo.class.getName() + " <filename>");
        System.exit(0);
    }
}
