/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.threads;

/**
 * Different Singleton realizations.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class SingletonDemo {

   public static void main(String[] args) {
      int iterations = 100000000;
      long start;
      synchronizedclass(iterations);
      holderclass(iterations);
      volatileinstance(iterations);
   }

   private static void synchronizedclass(int iterations) {
      long start;
      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) SynchronizedSingleton.getInstance();
      System.out.println("SynchronizedSingleton.getInstance(): " + (System.currentTimeMillis() - start));

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) SynchronizedSingleton.getInstance();
      System.out.println("SynchronizedSingleton.getInstance(): " + (System.currentTimeMillis() - start));
   }

   private static void holderclass(int iterations) {
      long start;
      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) HolderClassSingleton.getInstance();
      System.out.println("HolderClassSingleton.getInstance(): " + (System.currentTimeMillis() - start));

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) HolderClassSingleton.getInstance();
      System.out.println("HolderClassSingleton.getInstance(): " + (System.currentTimeMillis() - start));
   }

   private static void volatileinstance(int iterations) {
      long start;
      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) VolatileSingleton.getInstance();
      System.out.println("VolatileSingleton.getInstance(): " + (System.currentTimeMillis() - start));

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) VolatileSingleton.getInstance();
      System.out.println("VolatileSingleton.getInstance(): " + (System.currentTimeMillis() - start));
   }

   /**
    * Singleton that uses synchronization on the class object to make
    * sure that there is only one instance.
    */
   private static class SynchronizedSingleton {
      private static SynchronizedSingleton instance;

      private SynchronizedSingleton() {
      }

      public static synchronized SynchronizedSingleton
         getInstance() {
         if (instance == null) {
            instance = new SynchronizedSingleton();
         }
         return instance;
      }
   }

   /**
    * Singleton that uses the <i>initialize-on-demand holder class</i> idiom.
    */
   private static class HolderClassSingleton {
      /**
       * Holder class, that will only be initialized, when it
       * is used for the first time (JLS 12.4.1).
       */
      private static class SingletonHolder {
         static final HolderClassSingleton instance = new HolderClassSingleton();
      }

      private static SingletonHolder instanceHolder;

      private HolderClassSingleton() {
      }

      public static HolderClassSingleton getInstance() {
         return SingletonHolder.instance;
      }
   }

   /**
    * Singleton that uses a volatile instance variable.
    */
   private static class VolatileSingleton {
      private volatile static VolatileSingleton instance;

      private VolatileSingleton() {
      }

      public static VolatileSingleton getInstance() {
         if (instance == null) {
            instance = new VolatileSingleton();
         }
         return instance;
      }
   }
}
