/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.xml;

import com.tagtraum.perf.io.HuffmanOutputStream;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import java.io.*;
import java.util.zip.*;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.InputSource;

/**
 * Demonstrates the XML parsing of different formats.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class BSAXParseDemo {
   private static SAXParserFactory factory;

   public static void main(String[] args) throws Exception {
      File file = new File(args[0]);
      factory = SAXParserFactory.newInstance();
      //factory.setValidating(false);
      //factory.setNamespaceAware(false);
      File bsaxFile = createBSAXFile(file);
      File bsaxWoWsFile = createBSAXFileWithoutWhitespace(file);
      File gzipxmlFile = createGZIPXMLFile(file);

      System.out.println("Created test files...");
      System.gc();
      Thread.sleep(5000);

      long start;
      int iterations = 500;

      for (int j=0; j<3; j++) {
         runTests(iterations, bsaxFile, file, gzipxmlFile, bsaxWoWsFile);
      }
   }

   private static void runTests(int iterations, File bsaxFile, File file, File gzipxmlFile, File bsaxWoWsFile) throws Exception {
      long start;
      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseBSAXFile(bsaxWoWsFile);
      System.out.println("parseBSAXFileWithoutWhitespace: " + (System.currentTimeMillis()-start));

      System.gc();

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseBSAXFile(bsaxWoWsFile);
      System.out.println("parseBSAXFileWithoutWhitespace: " + (System.currentTimeMillis()-start));

      System.gc();

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseBSAXFile(bsaxFile);
      System.out.println("parseBSAXFile: " + (System.currentTimeMillis()-start));

      System.gc();

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseBSAXFile(bsaxFile);
      System.out.println("parseBSAXFile: " + (System.currentTimeMillis()-start));

      System.gc();

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseXMLFile(file);
      System.out.println("parseXMLFile: " + (System.currentTimeMillis()-start));

      System.gc();

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseXMLFile(file);
      System.out.println("parseXMLFile: " + (System.currentTimeMillis()-start));

      System.gc();

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseGZIPXMLFile(gzipxmlFile);
      System.out.println("parseGZIPXMLFile: " + (System.currentTimeMillis()-start));

      System.gc();

      start = System.currentTimeMillis();
      for (int i=0; i<iterations; i++) parseGZIPXMLFile(gzipxmlFile);
      System.out.println("parseGZIPXMLFile: " + (System.currentTimeMillis()-start));

      System.gc();
   }

   private static File createBSAXFile(File file) throws Exception {
      File out = File.createTempFile("test", "bsax");
      OutputStream outStream = new FileOutputStream(out);
      BSAXEncoder encoder = new BSAXEncoder(outStream);
      encoder.setIgnoreIgnorableWhitespace(false);
      SAXParser parser = factory.newSAXParser();
      parser.parse(file, encoder);
      outStream.close();
      return out;
   }

   private static File createBSAXFileWithoutWhitespace(File file) throws Exception {
      File out = File.createTempFile("test", "bsaxwows");
      OutputStream outStream = new FileOutputStream(out);
      BSAXEncoder encoder = new BSAXEncoder(outStream);
      encoder.setIgnoreIgnorableWhitespace(true);
      SAXParser parser = factory.newSAXParser();
      parser.parse(file, encoder);
      outStream.close();
      return out;
   }

   private static File createGZIPXMLFile(File file) throws Exception {
      File out = File.createTempFile("test", "xml.gz");
      GZIPOutputStream outStream = new GZIPOutputStream(new FileOutputStream(out));
      InputStream inStream = new FileInputStream(file);
      byte[] buf = new byte[8*1024];
      int len = 0;
      while ((len = inStream.read(buf)) != -1) {
         outStream.write(buf, 0, len);
      }
      outStream.finish();
      outStream.close();
      inStream.close();
      return out;
   }

   private static void parseBSAXFile(File file) throws Exception {
      InputSource in = new InputSource(new FileInputStream(file));
      BSAXDecoder decoder = new BSAXDecoder();
      DefaultHandler handler = new DefaultHandler();
      decoder.setContentHandler(handler);
      decoder.setDTDHandler(handler);
      decoder.parse(in);
   }

   private static void parseXMLFile(File file) throws Exception {
      SAXParser parser = factory.newSAXParser();
      parser.parse(file, new DefaultHandler());
   }

   private static void parseGZIPXMLFile(File file) throws Exception {
      GZIPInputStream in = new GZIPInputStream(new FileInputStream(file));
      SAXParser parser = factory.newSAXParser();
      parser.parse(in, new DefaultHandler());
      in.close();
   }
}

