/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.classloading;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Date;
import java.util.StringTokenizer;

/**
 * Launcher that allows multiple Java applications to run in the same VM.
 * Note that GUI applications usually will not be listed in the List field as they all run
 * in the same GUI-Thread and not in the thread they have been started in.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class MultiAppLauncher extends Frame implements ActionListener {

   private Label statusBar;
   private List appList;
   private TextField classPathField;
   private TextField mainClassField;
   private TextField argumentsField;

   public MultiAppLauncher() {
      super("MultiAppLauncher");
      addWindowListener(new WindowAdapter() {
         public void windowClosing(WindowEvent e) {
            dispose();
         }
      });
      classPathField = new TextField(20);
      mainClassField = new TextField(20);
      argumentsField = new TextField(20);
      Label classPathLabel = new Label("Classpath:");
      Label mainClassLabel = new Label("Main Class:");
      Label argumentsLabel = new Label("Arguments:");
      Button startButton = new Button("Start");
      startButton.addActionListener(this);
      appList = new List(10);
      statusBar = new Label();

      Panel dataPanel = new Panel();
      ((FlowLayout) dataPanel.getLayout()).setAlignment(FlowLayout.LEFT);
      dataPanel.add(mainClassLabel);
      dataPanel.add(mainClassField);
      dataPanel.add(classPathLabel);
      dataPanel.add(classPathField);
      dataPanel.add(argumentsLabel);
      dataPanel.add(argumentsField);

      Panel startButtonPanel = new Panel();
      ((FlowLayout) dataPanel.getLayout()).setAlignment(FlowLayout.RIGHT);
      startButtonPanel.add(startButton);

      Panel northPanel = new Panel();
      northPanel.add(dataPanel);
      northPanel.add(startButtonPanel);

      add(northPanel, BorderLayout.NORTH);
      add(appList, BorderLayout.CENTER);
      add(statusBar, BorderLayout.SOUTH);
      setSize(400, 400);
      pack();
      setVisible(true);
   }


   public void actionPerformed(ActionEvent e) {
      try {
         AppRunner appRunner = new AppRunner(classPathField.getText(), mainClassField.getText(), argumentsField.getText());
         appList.add(appRunner.getName());
         appRunner.start();
      } catch (Exception exception) {
         statusBar.setText(exception.toString());
      }
      repaint();
   }

   public static void main(String[] args) {
      new MultiAppLauncher();
   }

   private class AppRunner extends Thread {

      private final Class[] PARAMETERTYPES_FOR_MAIN = new Class[]{String[].class};

      private String[] args;
      private ClassLoader classLoader;
      private Class mainClass;
      private Method mainMethod;
      private String classPath;

      public AppRunner(String classPath, String mainClass, String args) throws ClassNotFoundException, MalformedURLException, NoSuchMethodException {
         super(new Date().toString() + ": java -classpath " + classPath + " " + mainClass + " " + args);
         this.args = createArgsArray(args);
         this.classPath = classPath;
         this.classLoader = createClassLoader(classPath);
         this.mainClass = this.classLoader.loadClass(mainClass);
         this.mainMethod = this.mainClass.getMethod("main", PARAMETERTYPES_FOR_MAIN);
      }

      public void run() {
         try {
            mainMethod.invoke(null, new Object[]{args});
            statusBar.setText(toString() + " exited.");
         } catch (Exception e) {
            System.err.println(this);
            e.printStackTrace();
            statusBar.setText(e.toString());
         }
         appList.remove(this.getName());
         repaint();
      }

      private ClassLoader createClassLoader(String classPath) throws MalformedURLException {
         StringTokenizer st = new StringTokenizer(classPath, File.pathSeparator);
         URL[] url = new URL[st.countTokens()];
         for (int i = 0; st.hasMoreTokens(); i++) {
            url[i] = new File(st.nextToken()).toURL();
         }
         // make sure we use the boot classloader for launching apps.
         return new URLClassLoader(url, String.class.getClassLoader());
      }

      private String[] createArgsArray(String args) {
         StringTokenizer st = new StringTokenizer(args, " \t\n\r");
         String[] argsArray = new String[st.countTokens()];
         for (int i = 0; st.hasMoreTokens(); i++) {
            argsArray[i] = st.nextToken();
         }
         return argsArray;
      }
   }
}
