/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * Base class for {@link Matrix} implementations.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public abstract class AbstractMatrix implements Matrix {

   private int cols;
   private int rows;
   private boolean growable;
   private static final String lf = System.getProperty("line.separator");
   private PropertyChangeSupport propertyChangeSupport;


   /**
    * Creates a matrix.
    *
    * @param cols number of cols this matrix should have.
    * @param rows number of rows this matrix should have.
    * @exception IllegalArgumentException if cols or rows is less than zero
    * @see #rows()
    * @see #columns()
    */
   public AbstractMatrix(int cols, int rows) {
      propertyChangeSupport = new PropertyChangeSupport(this);
      setColumns(cols);
      setRows(rows);
   }

   public void addPropertyChangeListener(PropertyChangeListener listener) {
      propertyChangeSupport.addPropertyChangeListener(listener);
   }

   public void removePropertyChangeListener(PropertyChangeListener listener) {
      propertyChangeSupport.removePropertyChangeListener(listener);
   }

   public boolean isGrowable() {
      return growable;
   }

   protected void setGrowable(boolean growable) {
      this.growable = growable;
   }

   public int columns() {
      return cols;
   }

   public int rows() {
      return rows;
   }

   protected void setColumns(int cols) {
      if (cols < 0) throw new IllegalArgumentException("cols = " + cols);
      propertyChangeSupport.firePropertyChange("matrix.columns", this.cols, cols);
      this.cols = cols;
   }

   protected void setRows(int rows) {
      if (rows < 0) throw new IllegalArgumentException("rows = " + rows);
      propertyChangeSupport.firePropertyChange("matrix.rows", this.cols, cols);
      this.rows = rows;
   }

   /**
    * To be called by subclasses to check whether col and row are in range.
    *
    * @param col column
    * @param row row
    * @exception IndexOutOfBoundsException if col or row are less than zero or greater than
    *    the size of this matrix
    * @see #rows()
    * @see #columns()
    */
   protected void rangeCheck(int col, int row) {
      if (growable) {
         if (col > columns()) setColumns(col);
         if (row > rows()) setRows(col);
      } else {
         if (col < 0 || col >= cols) throw new IndexOutOfBoundsException("col = " + col);
         if (row < 0 || row >= rows) throw new IndexOutOfBoundsException("row = " + row);
      }
   }

   /**
    * Dummy, better than nothing implementation for hashcode.
    * Needs to be overwritten for an efficient solution.
    */
   public int hashCode() {
      return toString().hashCode();
   }

   /**
    * Dummy, better than nothing implementation for hashcode.
    * Needs to be overwritten for an efficient solution.
    */
   public boolean equals(Object obj) {
      return toString().equals(obj.toString());
   }

   /**
    * Creates a string representation of this matrix.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer();
      for (int row = 0; row < rows(); row++) {
         for (int col = 0; col < columns(); col++) {
            sb.append(get(col, row));
            if (col != columns() - 1) sb.append("\t");
         }
         if (row != rows() - 1) sb.append(lf);
      }
      return sb.toString();
   }
}
