/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;
import java.util.BitSet;

/**
 * Pipe that sorts ints that lie within a certain range. I.e. offset &lt;= i &lt;= capacity.
 * After instantiation you are expected to provide a {@link ByteBuffer} to the
 * {@link #read(ByteBuffer)}-method. The ByteBuffer will be interpreted as an IntBuffer.
 * After you provided all int you wanted to sort, you may call the {@link #write(ByteBuffer)}-method
 * to receive the sorted ints. They will be written to the given {@link ByteBuffer}.
 * Note that even if you provided the same int twice, it will occur only once in the sorted
 * output.
 * <p>
 * This class is designed to work with {@link java.nio.channels.FileChannel#transferTo(long, long, WritableByteChannel)} and
 * {@link java.nio.channels.FileChannel#transferFrom(ReadableByteChannel, long, long)}. Note that some versions of Sun JDK 1.4.0
 * have bugs in these methods and cause the VM to crash.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see IntegerSortPipe1
 * @see java.nio.channels.FileChannel#transferTo(long, long, WritableByteChannel)
 * @see java.nio.channels.FileChannel#transferFrom(ReadableByteChannel, long, long)
 */
public class IntegerSortPipe implements ReadableByteChannel, WritableByteChannel {

   private boolean reading;
   private int offset;
   private int index;
   private BitSet bits;


   public IntegerSortPipe(int capacity, int offset) {
      this.offset = offset;
      bits = new BitSet(capacity);
   }

   /**
    * Call this method to read the sorted ints in a ByteBuffer.
    *
    * @param sortedInts ByteBuffer with sorted ints (after this method returns)
    * @return bytes written to the buffer
    *
    * @exception ClosedChannelException after you have started reading with
    *    {@link #read(ByteBuffer)} until you have read all sorted ints.
    */
   public int read(ByteBuffer sortedInts) throws IOException {
      if (!isOpen()) throw new ClosedChannelException();
      if (!reading) {
         reading = true;
         index = 0;
      }
      int pos = sortedInts.position();
      for (; index < bits.size() && sortedInts.remaining() > 3; index++) {
         if (bits.get(index)) sortedInts.putInt(index + offset);
      }
      int read = sortedInts.position() - pos;
      if (read == 0 && index == bits.size()) {
         reading = false;
         return -1;
      }
      return read;
   }

   /**
    * Call this method to write unsorted ints in a ByteBuffer.
    *
    * @param unsortedInts ByteBuffer with ints (4 bytes each)
    * @return bytes read from the buffer
    *
    * @exception ClosedChannelException after you have started reading with
    *    {@link #read(ByteBuffer)} until you have read all sorted ints.
    */
   public int write(ByteBuffer unsortedInts) throws IOException {
      if (!isOpen() || reading) throw new ClosedChannelException();
      int pos = unsortedInts.position();
      while (unsortedInts.remaining() > 3) {
         bits.set(unsortedInts.getInt() - offset);
      }
      return unsortedInts.position() - pos;
   }

   public boolean isOpen() {
      return bits != null;
   }

   public void close() throws IOException {
      if (bits == null) throw new ClosedChannelException();
      bits = null;
   }
}
