/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;

import java.util.Collections;
import java.util.Map;

/**
 * Caches a Map. Do not use this Cache for caching
 * a HashMap. You may get positive results for caching a TreeMap depending on the size
 * of the TreeMap,  the capacity of this cache and the quality of the hashcodes of the
 * cached objects.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class MapCache {
   private Map map;
   private Map unmodifiableMapView;
   private Cache cache;

   /**
    * Creates a cached map with an initial capacity that will be
    * increased to the next power of 2
    */
   public MapCache(Map map, Cache cache) {
      this.map = map;
      this.cache = cache;
      unmodifiableMapView = Collections.unmodifiableMap(map);
   }

   public int getCapacity() {
      return cache.getCapacity();
   }

   /**
    * Returns an unmodifiable view of the underlying map, so that modifications
    * of the map have to happen through the cached map.
    */
   public Map getMap() {
      return unmodifiableMapView;
   }

   public Object get(Object key) {
      Object value = cache.get(key);
      if (value != null) {
         return value;
      }
      value = map.get(key);
      if (key != null) {
         cache.put(key, value);
      }
      return value;
   }

   public Object put(Object key, Object value) {
      Object oldValue = map.put(key, value);
      if (key != null) {
         cache.put(key, value);
      }
      return oldValue;
   }

   public float getHitRatio() {
      return cache.getHitRatio();
   }

}
