/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;


/**
 * Operations for two dimensional <code>int</code> arrays.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class MatrixOperations {

   /**
    * Adds two int matrices.
    *
    * @param a matrix a
    * @param b matrix b
    * @return the resulting matrix
    */
   public static int[][] add(int[][] a, int[][] b) {
      int[][] c = new int[a.length][a[0].length];
      return add(a, b, c, 0, 0, 0, 0, 0, 0, a.length, a[0].length);
   }

   /**
    * Adds two int sub-matrices.
    *
    * @param a matrix a
    * @param b matrix b
    * @param c resulting matrix
    * @param ax starting row of the sub matrix in a that shall be used for the addition
    * @param ay starting column of the sub matrix in a that shall be used for the addition
    * @param bx starting row of the sub matrix in b that shall be used for the addition
    * @param by starting column of the sub matrix in b that shall be used for the addition
    * @param cx starting row of the sub matrix in c where the result shall be placed
    * @param cy starting column of the sub matrix in a where the result shall be placed
    * @param rows rows to add
    * @param columns columns to add
    * @return the resulting matrix
    * @see #add(int[][], int[][])
    */
   public static int[][] add(int[][] a, int[][] b, int[][] c, int ax, int ay, int bx, int by, int cx, int cy, int rows, int columns) {
      if (a == c) a = (int[][]) a.clone();
      if (b == c) b = (int[][]) b.clone();
      for (int x = 0; x < rows; x++) {
         int[] tc = c[cx + x];
         int[] ta = a[ax + x];
         int[] tb = b[bx + x];
         for (int y = 0; y < columns; y++) {
            tc[cy + y] = ta[ay + y] + tb[by + y];
         }
      }
      return c;
   }

   /**
    * Substracts two int matrices.
    *
    * @param a matrix a
    * @param b matrix b
    * @return the resulting matrix
    */
   public static int[][] subtract(int[][] a, int[][] b) {
      int[][] c = new int[a.length][a[0].length];
      return subtract(a, b, c, 0, 0, 0, 0, 0, 0, a.length, a[0].length);
   }

   /**
    * Substracts two int sub-matrices.
    *
    * @param a matrix a
    * @param b matrix b
    * @param c resulting matrix
    * @param ax starting row of the sub matrix in a that shall be used for the substraction
    * @param ay starting column of the sub matrix in a that shall be used for the substraction
    * @param bx starting row of the sub matrix in b that shall be used for the substraction
    * @param by starting column of the sub matrix in b that shall be used for the substraction
    * @param cx starting row of the sub matrix in c where the result shall be placed
    * @param cy starting column of the sub matrix in a where the result shall be placed
    * @param rows rows to substract
    * @param columns columns to substract
    * @return the resulting matrix
    * @see #add(int[][], int[][])
    */
   public static int[][] subtract(int[][] a, int[][] b, int[][] c, int ax, int ay, int bx, int by, int cx, int cy, int rows, int columns) {
      if (a == c) a = (int[][]) a.clone();
      if (b == c) b = (int[][]) b.clone();
      for (int x = 0; x < rows; x++) {
         int[] tc = c[cx + x];
         int[] ta = a[ax + x];
         int[] tb = b[bx + x];
         for (int y = 0; y < columns; y++) {
            tc[cy + y] = ta[ay + y] - tb[by + y];
         }
      }
      return c;
   }

   /**
    * Multiplies two int matrices. If this method is executed in a SMP
    * environment, the number of available processors times 16 threads
    * are started to compute the result concurrently.
    * This methods is synchronous.
    *
    * @param a matrix a
    * @param b matrix b
    * @return the resulting matrix
    */
   public static int[][] multiply(int[][] a, int[][] b) {
      int availableProcessors = Runtime.getRuntime().availableProcessors();
      if (availableProcessors == 1) {
         return conventionalMultiply(a, b);
      }
      return multithreadedMultiply(a, b, availableProcessors * 16);
   }

   static int[][] conventionalMultiply(int[][] a, int[][] b) {
      int[][] c = new int[a.length][b.length];
      for (int i = 0; i < a.length; i++) {
         int[] cTemp = c[i];
         for (int k = 0; k < b.length; k++) {
            int temp = 0;
            int[] aTemp = a[i];
            for (int n = 0; n < b.length; n++) {
               temp += aTemp[n] * b[n][k];
            }
            cTemp[k] = temp; // c[i][k] = temp
         }
      }
      return c;
   }

   /**
    * Multiplies two int matrices. Starts n threads
    * to compute the result concurrently.
    * This methods is synchronous.
    *
    * @param a matrix a
    * @param b matrix b
    * @param threads the number of threads to start. If threads is &lt;=1 no separate thread
    *    is started
    * @return the resulting matrix
    */
   public static int[][] multithreadedMultiply(int[][] a, int[][] b, int threads) {
      if (threads <= 1) {
         return conventionalMultiply(a, b);
      }
      int[][] c = new int[a.length][b.length];
      try {
         Thread[] t = new Thread[threads];
         int length = (int) Math.ceil(a.length / (double) threads);
         for (int i = 0; i < threads; i++) {
            t[i] = new Thread(new MultiprocessorMultiplier(a, b, c, length * i, length));
            t[i].start();
         }
         for (int i = 0; i < threads; i++) {
            t[i].join();
         }
      } catch (InterruptedException ie) {
         ie.printStackTrace();
      }
      return c;
   }

   private static class MultiprocessorMultiplier implements Runnable {
      private int[][] a;
      private int[][] b;
      private int[][] c;
      private int startRow;
      private int length;

      public MultiprocessorMultiplier(int[][] a, int[][] b, int[][] c, int startRow, int length) {
         this.a = a;
         this.b = b;
         this.c = c;
         this.startRow = startRow;
         this.length = length;
      }

      public void run() {
         for (int i = startRow, len = length + startRow; i < len && i < a.length; i++) {
            int[] cTemp = c[i];
            for (int k = 0; k < b.length; k++) {
               int temp = 0;
               int[] aTemp = a[i];
               for (int n = 0; n < b.length; n++) {
                  temp += aTemp[n] * b[n][k];
               }
               cTemp[k] = temp;
            }
         }
      }
   }

}
