/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import java.io.*;

/**
 * {@link FileCache.Entry} object that uses a byte array to cache the content.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
class ByteArrayFile implements FileCache.Entry {
   private byte[] content;
   private File file;
   private long length;
   private long lastModified;
   private int timeToLive;
   private long creationTime;
   private long lastValidityCheck;

   /**
    * Initializes this entry with a file, that is read into a byte array.
    */
   public ByteArrayFile(File file) throws IOException {
      this.file = file;
      load(file);
      creationTime = System.currentTimeMillis();
      lastValidityCheck = creationTime;
   }

   /**
    * Initializes this entry with a file, that is read into a byte array.
    */
   private void load(File file) throws IOException {
      this.length = file.length();
      this.lastModified = file.lastModified();
      content = new byte[(int)length];
      InputStream in = new FileInputStream(file);
      try {
         int offset = 0;
         while (offset < length) {
            offset += in.read(content, offset, (int)length-offset);
         }
      }
      finally {
         if (in != null) try { in.close(); } catch (IOException ioe) {}
      }
   }

   public InputStream getNewInputStream() throws IOException {
      return new ByteArrayInputStream(content);
   }

   public File getFile() {
      return file;
   }

   public long length() {
      return length;
   }

   public long lastModified() {
      return lastModified;
   }

   public void setTimeToLive(int timeToLive) {
      this.timeToLive = timeToLive;
   }

   public int getTimeToLive() {
      return timeToLive;
   }

   public boolean isStale() {
      boolean stale = false;
      long now = System.currentTimeMillis();
      if (timeToLive >= 0 && lastValidityCheck + (long)timeToLive < now) {
         // time is up, but maybe the thing is still not stale
         stale = length != file.length() || lastModified != file.lastModified();
         if (!stale) lastValidityCheck = now;
      }
      return stale;
   }

}
