/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import com.tagtraum.perf.util.HuffmanTree;
import com.tagtraum.perf.util.HuffmanException;

import java.io.*;

/**
 * Implementation of a HuffmanInputStream. <em>Note that this implementation is not
 * optimized.</em>
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see HuffmanOutputStream
 * @see HuffmanTree
 */
public class HuffmanInputStream extends FilterInputStream {

   private byte[] buffer = new byte[0];
   private int pos;
   private HuffmanTree huffman;
   private static byte[] skipBuffer = new byte[1024];

   public HuffmanInputStream(InputStream in) {
      super(in);
      this.huffman = new HuffmanTree();
   }

   /**
    * Reads a Huffman encoded block.
    */
   private void readBlock() throws IOException {
      buffer = huffman.decode(in);
      pos = 0;
   }

   /**
    * @exception HuffmanException if the stream is corrupted and cannot be decoded properly. This might
    *    be the case if the stream was not able to read a complete block.
    */
   public int read() throws IOException, HuffmanException {
      if (pos == buffer.length) {
         try {
            readBlock();
         }
         catch (EOFException eof) {
            return -1;
         }
      }
      return buffer[pos++];
   }

   /**
    * @exception HuffmanException if the stream is corrupted and cannot be decoded properly. This might
    *    be the case if the stream was not able to read a complete block.
    */
   public int read(byte b[], int off, int len) throws IOException, HuffmanException {
      int read = 0;
      while (available() < len) {
         System.arraycopy(buffer, pos, b, off + read, available());
         read += available();
         try {
            readBlock();
         }
         catch (EOFException eof) {
            return read == 0 ? -1 : read;
         }
      }
      System.arraycopy(buffer, pos, b, off + read, len - read);
      pos += len;
      return len;
   }

   public long skip(long n) throws IOException {
      for (int skipped=0; skipped<n; ) {
         int read = read(skipBuffer);
         if (read<=0) return skipped;
         skipped += read;
      }
      return n;
   }

   public boolean markSupported() {
      return false;
   }

   public void mark(int readlimit) {
   }

   public void reset() throws IOException {
   }

   public int available() throws IOException {
      return buffer.length-pos;
   }
}
