/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import java.io.CharConversionException;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.util.ResourceBundle;


/**
 * Writes only bytes that fit through a defined mask.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class MaskedStreamWriter extends Writer {

   public static final int ASCII_MASK = 0x7f;
   public static final int ISO_8859_1_MASK = 0xff;

   private OutputStream out;
   private byte[] buf;
   private static ResourceBundle localStrings = ResourceBundle.getBundle("com.tagtraum.perf.io.localStrings");
   private int mask;

   /**
    * @param mask mask has to be less than 256
    * @param out the OutputStream to write to.
    */
   public MaskedStreamWriter(OutputStream out, int mask) {
      this(out, 8 * 1024, mask); // 8kb
   }

   /**
    * @param mask mask has to be less than 256
    * @param out the OutputStream to write to.
    * @param bufsize size of the interim buffer bytes are written to
    */
   public MaskedStreamWriter(OutputStream out, int bufsize, int mask) {
      this.out = out;
      buf = new byte[bufsize]; // 8kb
      if (mask > 0xff) throw new IllegalArgumentException(localStrings.getString("illegal_mask"));
      this.mask = mask;
   }

   public void write(int c) throws IOException {
      if ((c & ~mask) != 0) throw new CharConversionException(localStrings.getString("unallowed_char") + " " + c);
      out.write(c);
   }

   public void write(char[] cbuf, int offset, int length) throws IOException {
      int l = Math.min(buf.length, length);
      while (l > 0) {
         int end = offset + l;
         for (int i = offset; i < end; i++) {
            char c = cbuf[i];
            if ((c & ~mask) != 0)
               throw new CharConversionException(localStrings.getString("unallowed_char") + " " + c);
            else
               buf[i - offset] = (byte) c;
         }
         out.write(buf, 0, l);
         offset += l;
         length -= l;
         l = Math.min(buf.length, length);
      }
   }

   public void write(String str, int off, int len) throws IOException {
      if (len < 0)
         throw new IndexOutOfBoundsException();
      char cbuf[] = new char[len];
      str.getChars(off, off + len, cbuf, 0);
      write(cbuf, 0, len);
   }


   public void close() throws IOException {
      out.close();
   }

   public void flush() throws IOException {
      out.flush();
   }

}
