/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import java.io.*;

/**
 * Stream that can read bits und halfbytes.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see XtendedDataOutputStream
 */
public class XtendedDataInputStream extends DataInputStream {

   private int bitNumber = 8;
   private byte currentBitByte;

   private boolean firsthalf = true;
   private int currentHalfByte;

   private InputStream unfilteredIn;

   public XtendedDataInputStream(InputStream in) {
      super(new DiscardingFilterInputStream(in));
      ((DiscardingFilterInputStream)this.in).setXtendedDataInputStream(this);
      this.unfilteredIn = in;
   }

   /**
    * Reads a bit from the underlying stream.
    * <em>Note:</em> If you are trying to read more bits than were
    * originally written, you will always get values for a multiple
    * of 8 bits. I.e. if you have written 4 bits, you will nevertheless
    * be able to read 8 bits. The bits not explicitly written will
    * default to false when read.
    *
    * @return true for 1 and false for 0
    * @exception EOFException when the end of this stream has been reached.
    */
   public boolean readBit() throws IOException {
      if (bitNumber == 8) {
         bitNumber = 0;
         int i = unfilteredIn.read();
         if (i == -1) throw new EOFException();
         currentBitByte = (byte)i;
      }
      return ((1 << bitNumber++) & currentBitByte) != 0;
   }

   /**
    * Reads half a byte from the underlying stream.
    * <em>Note:</em> If you are trying to read more half bytes than were
    * originally written, you will always get values for a multiple
    * of bytes (not half bytes). I.e. if you have written half a byte, you will nevertheless
    * be able to read a full byte. The bits not explicitly written will
    * default to false when read.
    *
    * @return half a byte
    * @exception EOFException when the end of this stream has been reached.
    */
   public int readHalfByte() throws IOException {
      bitNumber = 8;
      if (firsthalf) {
         firsthalf = false;
         currentHalfByte = unfilteredIn.read();
         if (currentHalfByte == -1) return -1;
         return ((byte)currentHalfByte) & 0x0f;
      }
      else {
         firsthalf = true;
         return (currentHalfByte>>4) & 0x0f;
      }
   }

   private void discardBits() {
      bitNumber = 8;
   }

   private void discardHalfByte() {
      firsthalf = true;
   }

   /**
    * If less than a full byte was read, calling this method discards the remaining
    * bits of the current byte.
    */
   public void discard() throws IOException {
      discardBits();
      discardHalfByte();
   }

   // this is to work around the unpleseant fact that all of DataInputStreams methods are final.
   private static class DiscardingFilterInputStream extends FilterInputStream {
      private XtendedDataInputStream xin;
      public DiscardingFilterInputStream(InputStream in) {
         super(in);
      }

      private void setXtendedDataInputStream(XtendedDataInputStream xin) {
         this.xin = xin;
      }

      public int read() throws IOException {
         xin.discardBits();
         if (!xin.firsthalf) {
            int c = xin.readHalfByte();
            c |= xin.readHalfByte() << 4;
            return c;
         }
         else {
            return super.read();
         }
      }

      public int read(byte b[], int off, int len) throws IOException {
         xin.discardBits();
         if (!xin.firsthalf) {
            int count = 0;
            for (int i=off; i<off + len; i++) {
               int c = xin.readHalfByte();
               if (c == -1) return count;
               int d = xin.readHalfByte();
               if (d == -1) return count;
               count++;
               b[i] = (byte)(c | (d <<4));
            }
            return count;
         }
         else {
            return super.read(b, off, len);
         }
      }

      public long skip(long n) throws IOException {
         xin.discardBits();
         return super.skip(n);
      }

      public boolean markSupported() {
         return false;
      }
   }

}
