/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.strings;

import java.text.DateFormat;
import java.text.FieldPosition;
import java.text.ParsePosition;
import java.util.Date;

/**
 * DateFormat that prints the difference to the first time it is used to output a date.
 * Note: Only the format methods are functional!
 *
 * Date: Feb 23, 2002
 * Time: 5:44:23 PM
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class DifferenceDateFormat extends DateFormat {

    private long baseTime;

    public DifferenceDateFormat() {
    }

    /**
     * The first time this method is called, the long value of the date is appended to
     * the StringBuffer. Every following time, the difference to the first time is appended.
     * Positive values indicate dates <i>after</i> the base date.<p>
     * Note that format and parse share the same base time.
     */
    public synchronized StringBuffer format(Date date, StringBuffer toAppendTo,
                                            FieldPosition fieldPosition) {
        if (baseTime == 0) {
            baseTime = date.getTime();
            toAppendTo.append(Long.toString(baseTime));
        } else {
            long diff = date.getTime() - baseTime;
            // use more efficient Integer.toString() method, if possible
            if (diff <= Integer.MAX_VALUE && diff >= Integer.MIN_VALUE)
                toAppendTo.append((int) diff);
            else
                toAppendTo.append(diff);
        }
        return toAppendTo;
    }

    /**
     * Parses a String. The first time this method is called, the String is taken as
     * base time in ms from 1970 for all following calls. Those will be added/subtracted from the
     * base time.<p>
     * Note that format and parse share the same base time.
     */
    public Date parse(String text, ParsePosition pos) {
        int max = text.length();
        long result = 0;
        int digit = 0;
        int length = 0;
        long limit = 0;
        int i = pos.getIndex();
        boolean negative = false;
        int maxChars = 0;

        if (text.charAt(i) == '-') {
            // shortcut for ints
            negative = true;
            limit = Long.MIN_VALUE;
            maxChars = NumberParser.MAX_NEGATIVE_LONG_CHARS;
            i++;
        } else {
            limit = -Long.MAX_VALUE;
            maxChars = NumberParser.MAX_POSITIVE_LONG_CHARS;
        }

        while(i<text.length()) {
            digit = text.charAt(i)-'0';
            if (digit < 0 || digit > 9) {
                if (length == 0) {
                    pos.setErrorIndex(i);
                    return null;
                }
                else {
                    break;
                }
            }
            length++;
            if (length > maxChars) {
                pos.setErrorIndex(i);
                return null;
            }
            result *= 10;
            if (result < limit + digit) {
                pos.setErrorIndex(i);
                return null;
            }
            result -= digit;
            i++;
        }
        pos.setIndex(i);
        synchronized (this) {
            if (baseTime == 0) {
                baseTime = negative ? result:-result;
                return new Date(baseTime);
            }
            else {
                return new Date(baseTime + (negative ? result : -result));
            }
        }
    }

}
