/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.threads;

import java.util.Arrays;

/**
 * Instances of this class are able to execute the <code>run()</code> method of a given
 * {@link java.lang.Runnable} in a thread. Threads are pooled and reused.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class Runner {

   private LimitedStack stack;
   private Class runnerThreadClass;
   private boolean destroyed;
   private int capacity;

   public Runner(int capacity) {
      if (capacity <= 0) throw new IllegalArgumentException("Capacity must be greater than zero.");
      this.capacity = capacity;
      this.stack = new LimitedStack(capacity);
      setRunnerThreadClass(RunnerThread.class);
   }

   /**
    * Runs the runnable in a thread.
    *
    * @param runnable a runnable that shall be run
    */
   public void run(Runnable runnable) {
      getRunnerThread(runnable).run();
   }

   protected synchronized RunnerThread getRunnerThread(Runnable runnable) {
      if (isDestroyed()) throw new IllegalStateException("Runner is destroyed.");
      RunnerThread rt = (RunnerThread) stack.pop();
      if (rt == null) {
         rt = newRunnerThread();
         rt.setRunner(this);
      }
      rt.setRunnable(runnable);
      return rt;
   }

   protected RunnerThread newRunnerThread() {
      try {
         return (RunnerThread) getRunnerThreadClass().newInstance();
      } catch (Exception e) {
         throw new InternalError(e.toString());
      }
   }

   /**
    * RunnerThread class.
    */
   public Class getRunnerThreadClass() {
      return runnerThreadClass;
   }

   protected void setRunnerThreadClass(Class runnerThreadClass) {
      this.runnerThreadClass = runnerThreadClass;
   }

   synchronized boolean returnRunnerThread(RunnerThread rt) {
      return rt.isAlive() && !destroyed && stack.push(rt);
   }

   /**
    * Number of RunnerThreads that the pool can hold.
    */
   public int getCapacity() {
      return capacity;
   }

   /**
    * Number of RunnerThreads currently in the pool.
    */
   public int size() {
      return stack.size();
   }

   public String toString() {
      return "Runner " + stack + " [destroyed: " + destroyed + "]";
   }

   /**
    * Indicates whether this Runner is destroyed.
    */
   public synchronized boolean isDestroyed() {
      return destroyed;
   }

   /**
    * Stops and dereferences all RunnerThreads and marks this Runner as destroyed.
    */
   public synchronized void destroy() {
      if (destroyed) throw new IllegalStateException("Runner is already destroyed.");
      clear();
      destroyed = true;
   }

   /**
    * Stops and dereferences all RunnerThreads.
    */
   public synchronized void clear() {
      for (RunnerThread rt = stack.pop(); rt != null; rt = stack.pop()) {
         rt.quit();
      }
   }

   protected void finalize() throws Throwable {
      if (!destroyed) destroy();
      super.finalize();
   }

   private static class LimitedStack {
      private RunnerThread[] values;
      private int size;

      public LimitedStack(int capacity) {
         values = new RunnerThread[capacity];
      }

      public RunnerThread pop() {
         if (size == 0) return null;
         RunnerThread rt = values[--size];
         values[--size] = null;
         return rt;
      }

      public boolean push(RunnerThread rt) {
         if (size == values.length - 1) return false;
         values[size++] = rt;
         return true;
      }

      public int size() {
         return size;
      }

      public String toString() {
         return "[size: " + size + ", capacity: " + values.length + ", elements: " + Arrays.asList(values).toString() + "]";
      }
   }
}
