/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.xml;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;
import java.io.*;
import java.util.*;

import com.tagtraum.perf.io.XtendedDataOutputStream;

/**
 * Encodes SAX-Events in a binary format.<p>
 * This class is not threadsafe.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see BSAXDecoder
 */
public class BSAXEncoder extends DefaultHandler {

   static final int SWITCH_CODE_PAGE = 0;
   static final int END = 1;

   static final int NEW_BLOCK = 2;
   static final int START_DOCUMENT = 3;
   static final int END_DOCUMENT = 4;
   static final int START_ELEMENT = 5;
   static final int END_ELEMENT = 6;
   static final int CHARACTERS = 7;
   static final int IGNORABLE_WHITESPACE = 8;
   static final int PROCESSING_INSTRUCTION = 9;
   static final int SKIPPED_ENTITY = 10;
   static final int START_PREFIXMAPPING = 11;
   static final int END_PREFIXMAPPING = 12;
   static final int NOTATION_DECLARATION = 13;
   static final int UNPARSED_ENTITY_DECLARATION = 14;

   static final int RESERVED_CODES = 2;
   static final int CODES_PER_CODEPAGE = 16;
   static final int DEFINABLE_CODES_PER_CODEPAGE = CODES_PER_CODEPAGE-RESERVED_CODES;

   private static int DEFAULT_BLOCKSIZE = 8*1024;
   private static ResourceBundle localStrings = ResourceBundle.getBundle("com.tagtraum.perf.xml.localStrings");

   private XtendedDataOutputStream out;
   private XtendedDataOutputStream bufferedDataStream;
   private ByteArrayOutputStream buffer;
   private int currentCodePage;
   private Map symbolTable;
   private List symbolList;
   private Map oldSymbolTable;
   private List oldSymbolList;
   private int code = RESERVED_CODES;
   private int codePage;
   private int blockSize;
   private boolean firstFlush;
   private StringBuffer characters;
   private boolean ignoreIgnorableWhitespace;

   /**
    * Sets the outputstream and the default blocksize of 8k.
    */
   public BSAXEncoder(OutputStream out) {
      this(out, DEFAULT_BLOCKSIZE);
   }

   /**
    * Sets the outputstream and the blocksize. Blocksize corresponds to the size of
    * the internally used buffer.
    */
   public BSAXEncoder(OutputStream out, int blockSize) {
      this.out = out instanceof BufferedOutputStream ? new XtendedDataOutputStream(out) : new XtendedDataOutputStream(new BufferedOutputStream(out));
      this.blockSize = blockSize;
      buffer = new ByteArrayOutputStream(blockSize + 256); // some extra space
      bufferedDataStream = new XtendedDataOutputStream(buffer);
      symbolTable = new HashMap();
      symbolList = new ArrayList();
      oldSymbolTable = new HashMap();
      oldSymbolList = new ArrayList();
      firstFlush = true;
      characters = new StringBuffer(1024);
   }

   /**
    * @see #setIgnoreIgnorableWhitespace(boolean)
    */
   public boolean isIgnoreIgnorableWhitespace() {
      return ignoreIgnorableWhitespace;
   }

   /**
    * Set this flag to true, if you want to ignore 'ignorable whitespace', i.e.
    * whitespace that appears in element content as defined in XML 1.0 3.2.1.
    * Setting this flag to true potentially reduces the size of the document dramatically,
    * if you are using a validating parser.
    * <p>
    * By default this is set to <code>false</code>
    *
    * @see #isIgnoreIgnorableWhitespace()
    */
   public void setIgnoreIgnorableWhitespace(boolean ignoreIgnorableWhitespace) {
      this.ignoreIgnorableWhitespace = ignoreIgnorableWhitespace;
   }

   private void writeVersion() throws IOException {
      // version 1.0
      out.write(1); // major
      out.write(0); // minor
   }

   private void checkBufferSize() throws IOException {
      if (buffer.size() > blockSize) {
         flush();
      }
   }

   private void flush() throws IOException {
      if (firstFlush) {
         writeVersion();
         firstFlush = false;
      }

      bufferedDataStream.flush();
      writeSymbolTable();
      out.write(buffer.toByteArray());
      out.flush();
      buffer.reset();
   }

   public void startElement(String uri, String localName,
                            String qName, Attributes attributes)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(START_ELEMENT);
         writeSymbol(getSymbol(qName));
         if (attributes != null) {
            for (int i = 0, len =attributes.getLength(); i < len; i++) {
               writeSymbol(getSymbol(attributes.getQName(i)));
               writeSymbol(getSymbol(attributes.getType(i)));
               bufferedDataStream.writeUTF(attributes.getValue(i));
            }
         }
         writeCode(END);
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void endElement(String uri, String localName, String qName)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(END_ELEMENT);
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void startDocument()
         throws SAXException {
      flushCharacters();
      try {
         writeCode(START_DOCUMENT);
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void endDocument()
         throws SAXException {
      flushCharacters();
      try {
         writeCode(END_DOCUMENT);
         // no buffersize checking as this is the last call anyway.
         flush();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void characters(char ch[], int start, int length)
         throws SAXException {
      if (length > 0) {
         characters.append(ch, start, length);
      }
   }

   public void flushCharacters() throws SAXException {
      if (characters.length() > 0) {
         try {
            writeCode(CHARACTERS);
            bufferedDataStream.writeUTF(characters.toString());
            checkBufferSize();
            // reset the StringBuffer and
            // make sure that it does not stay to large
            if (characters.length() > 8*1024) characters = new StringBuffer(1024);
            else characters.setLength(0);
         } catch (IOException ioe) {
            throw new SAXException(localStrings.getString("exception_during_io"), ioe);
         } catch (RuntimeException re) {
            throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
         }
      }
   }

   public void ignorableWhitespace(char ch[], int start, int length)
         throws SAXException {
      flushCharacters();
      if (!ignoreIgnorableWhitespace && length > 0) {
         try {
            writeCode(IGNORABLE_WHITESPACE);
            bufferedDataStream.writeUTF(new String(ch, start, length));
            checkBufferSize();
         } catch (IOException ioe) {
            throw new SAXException(localStrings.getString("exception_during_io"), ioe);
         } catch (RuntimeException re) {
            throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
         }
      }
   }

   public void processingInstruction(String target, String data)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(PROCESSING_INSTRUCTION);
         bufferedDataStream.writeUTF(target);
         bufferedDataStream.writeUTF(data);
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void skippedEntity(String name)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(SKIPPED_ENTITY);
         writeSymbol(getSymbol(name));
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void startPrefixMapping(String prefix, String uri)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(START_PREFIXMAPPING);
         writeSymbol(getSymbol(prefix));
         writeSymbol(getSymbol(uri));
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void endPrefixMapping(String prefix)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(END_PREFIXMAPPING);
         writeSymbol(getSymbol(prefix));
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void notationDecl(String name, String publicId, String systemId)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(NOTATION_DECLARATION);
         bufferedDataStream.writeUTF(name);
         bufferedDataStream.writeUTF(publicId);
         bufferedDataStream.writeUTF(systemId);
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void unparsedEntityDecl(String name, String publicId,
                                  String systemId, String notationName)
         throws SAXException {
      flushCharacters();
      try {
         writeCode(UNPARSED_ENTITY_DECLARATION);
         bufferedDataStream.writeUTF(name);
         bufferedDataStream.writeUTF(publicId);
         bufferedDataStream.writeUTF(systemId);
         bufferedDataStream.writeUTF(notationName);
         checkBufferSize();
      } catch (IOException ioe) {
         throw new SAXException(localStrings.getString("exception_during_io"), ioe);
      } catch (RuntimeException re) {
         throw new SAXException(localStrings.getString("unexpected_exception_during_io"), re);
      }
   }

   public void fatalError(SAXParseException e)
         throws SAXException {
      throw e;
   }

   private Symbol getSymbol(String string) {
      Symbol s = (Symbol) symbolTable.get(string);
      if (s == null) s = (Symbol) oldSymbolTable.get(string);
      if (s == null) {
         if (code == CODES_PER_CODEPAGE) {
            codePage++;
            code = RESERVED_CODES;
         }
         s = new Symbol(string, code, codePage);
         symbolTable.put(string, s);
         symbolList.add(string);
         code++;
      }
      return s;
   }

   private void writeCode(int code) throws IOException {
      //System.out.println("Encoder Code: " + code);
      bufferedDataStream.writeHalfByte(code);
   }

   private void writeSymbol(Symbol s) throws IOException {
      if (currentCodePage != s.getCodePage()) {
         bufferedDataStream.writeHalfByte(SWITCH_CODE_PAGE);
         bufferedDataStream.writeHalfByte(s.getCodePage() & 0x0f);
         bufferedDataStream.writeHalfByte(s.getCodePage() & 0xf0);
         currentCodePage = s.getCodePage();
      }
      bufferedDataStream.writeHalfByte(s.getCode());
   }

   private void writeSymbolTable() throws IOException {
      ByteArrayOutputStream bout = new ByteArrayOutputStream();
      XtendedDataOutputStream symbolTableStream = new XtendedDataOutputStream(bout);
      symbolTableStream.writeHalfByte(NEW_BLOCK);
      symbolTableStream.writeInt(symbolTable.size());
      for (int i = 0, c = symbolList.size(); i < c; i++) {
         symbolTableStream.writeUTF((String) symbolList.get(i));
      }
      symbolTableStream.flush();
      out.write(bout.toByteArray());

      symbolList.clear();
      oldSymbolTable.putAll(symbolTable);
      symbolTable.clear();
   }


   /**
    * Encodes a given XML file. The file is validated and the parser for reading the file
    * is aware of namespaces. Ignorable whitespace will be ignored - i.e. is not part of
    * the encoded file.
    */
   public static void main(String[] args) throws IOException, ParserConfigurationException, SAXException {
      SAXParserFactory factory = SAXParserFactory.newInstance();
      factory.setValidating(true);
      factory.setNamespaceAware(true);
      SAXParser parser = factory.newSAXParser();
      String fileName = args[0];
      String compressedFileName = fileName.substring(0, fileName.lastIndexOf('.')) + ".bsax";
      OutputStream out = null;
      try {
         out = new FileOutputStream(compressedFileName);
         BSAXEncoder encoder = new BSAXEncoder(out);
         encoder.setIgnoreIgnorableWhitespace(true);
         parser.parse(new File(fileName), encoder);
      }
      finally {
         if (out != null) try {out.close();} catch(IOException ioe) {}
      }
   }

   private static class Symbol {
      private int code;
      private int codePage;
      private String value;

      public Symbol(String value, int code, int codePage) {
         this.value = value;
         this.code = code;
         this.codePage = codePage;
      }

      public int getCode() {
         return code;
      }

      public int getCodePage() {
         return codePage;
      }

      public String getValue() {
         return value;
      }

      public String toString() {
         return codePage + ":" + code + " " + value;
      }

   }
}
